"""
DSLIBRARY - run your data science code anywhere.

"""
__version__ = "0.0.27"

# environment variables
from .front import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TOKEN
# main base class
from .front import DSLibrary
# run models
from .run_models import ModelRunner
# utils for doing setup
from .utils.instantiation import parse_verb_and_args, instantiate_class_and_args

# if a package called 'mmlibrary' is installed it becomes the default target
try:
    import mmlibrary
except ImportError:
    mmlibrary = None


def instance(env: dict=None):
    """
    Get an instance of mmlibrary methods.  Configuration is through environment variables.

    To run locally in the current folder:   (this is the default)
      DSLIBRARY_TARGET=local

    To run locally with a specific default folder:
      DSLIBRARY_TARGET=local:<folder>

    To run against a REST API:
      DSLIBRARY_TARGET=https://hostname:port/path/to/api/
      DSLIBRARY_TOKEN=<access token, or credentials>

    To use a custom implementation:
      DSLIBRARY_TARGET=package.ClassName:value:name=value

    Parameter values and data input/output locations/formats and such are defined by setting this variable to
    JSON (see ModelRunner).
      DSLIBRARY_SPEC={...}

    :param env:  Instead of using os.environ, environment values can be specified here.

    :returns:    An instance of the MMLibrary() base class.
    """
    from .transport.to_local import DSLibraryLocal
    from .transport.to_rest import DSLibraryViaREST
    from .transport.to_volume import DSLibraryViaVolume
    from .transport.to_mmlibrary import DSLibraryViaMMLibrary

    import json
    import os
    if env is None:
        env = os.environ
    spec = env.get(ENV_DSLIBRARY_SPEC) or "{}"
    if isinstance(spec, str):
        spec = json.loads(spec)
    target = env.get(ENV_DSLIBRARY_TARGET) or ("mmlibrary" if mmlibrary else "local")
    target_parts = target.split(":")
    # point to a REST API
    if target_parts[0] in ("http", "https"):
        token = env.get(ENV_DSLIBRARY_TOKEN)
        return DSLibraryViaREST(url=target, token=token)
    # FIXME protocol handlers need to use 'env'
    # gather arguments
    mode, args, kwargs = parse_verb_and_args(target)
    if spec:
        kwargs["spec"] = spec
    if env is not None:
        kwargs["_env"] = env
    # a strictly local instance
    if target_parts[0] == "local":
        return DSLibraryLocal(*args, **kwargs)
    # delegation to 'mmlibrary'
    if target_parts[0] == "mmlibrary":
        return DSLibraryViaMMLibrary(*args, **kwargs)
    # writes through a shared volume, i.e. so that a sidecar can perform communications for us
    if target_parts[0] == "volume":
        return DSLibraryViaVolume(*args, **kwargs)
    # we fall through to custom implementation support
    return instantiate_class_and_args(target, DSLibrary)


def run_model_method(method, dsl=None):
    """
    Call a method representing a model, passing along the parameters that come from either a supplied dslibrary instance
    (dsl), or from the default dslibrary instance.

    :param method:  Method to call.
    :param dsl:     Dslibrary instance which will fetch the calling parameters.
    :return:        Return value from the method (or captured KPIs if that is configured).
    """
    params = dsl.get_parameters() if dsl else get_parameters()
    return ModelRunner(parameters=params).run_method(method)


# set up default methods
_DEFAULT = instance()
get_parameter = _DEFAULT.get_parameter
get_parameters = _DEFAULT.get_parameters
load_dataframe = _DEFAULT.load_dataframe
open_resource = _DEFAULT.open_resource
write_resource = _DEFAULT.write_resource
log_metric = _DEFAULT.log_metric
log_metrics = _DEFAULT.log_metrics
log_param = _DEFAULT.log_param
log_artifact = _DEFAULT.log_artifact
log_artifacts = _DEFAULT.log_artifacts
log_dict = _DEFAULT.log_dict
log_text = _DEFAULT.log_text
get_sql_connection = _DEFAULT.get_sql_connection
open_run_data = _DEFAULT.open_run_data
read_run_data = _DEFAULT.read_run_data
write_run_data = _DEFAULT.write_run_data
set_evaluation_result = _DEFAULT.set_evaluation_result
open_model_binary = _DEFAULT.open_model_binary
load_pickled_model = _DEFAULT.load_pickled_model
save_pickled_model = _DEFAULT.save_pickled_model
read_resource = _DEFAULT.read_resource
start_run = _DEFAULT.start_run
end_run = _DEFAULT.end_run
active_run = _DEFAULT.active_run
get_metadata = _DEFAULT.get_metadata
get_uri = _DEFAULT.get_uri
get_last_metric = _DEFAULT.get_last_metric
get_filesystem_connection = _DEFAULT.get_filesystem_connection
install_packages = _DEFAULT.install_packages
scoring_response = _DEFAULT.scoring_response
# TODO there may be more methods to add here



import os
import io
import unittest

from dslibrary import DSLibrary, instance
from dslibrary.proto_handlers import load_protocol_handler, ProtocolHandler


class TestProtoHandlers(unittest.TestCase):

    def test_load_protocol_handler(self):
        """
        The basic idea: set an environment variable telling us how to handle a given protocol.
        """
        os.environ["DSLIBRARY_PROTO_ZZZ"] = "tests.test_proto_handlers.MyProtoZZZ:v=hello"
        handler = load_protocol_handler("zzz://host/etc")
        assert handler
        assert handler.open_resource("x", "rb").read() == b'hello'

    def test_through_open_resource(self):
        os.environ["DSLIBRARY_PROTO_ZZZ"] = "tests.test_proto_handlers.MyProtoZZZ:v=hello2"
        with DSLibrary().open_resource("x", "rb", uri="zzz://x") as f:
            assert f.read() == b'hello2'

    def test_through_get_filesystem(self):
        os.environ["DSLIBRARY_PROTO_ZZZ"] = "tests.test_proto_handlers.MyProtoZZZ:v=hello2"
        fs = DSLibrary().get_filesystem_connection("zzz://x")
        with fs.open("x") as f_r:
            assert f_r.read() == b'from_fs'

    def test_through_dslibrary_instance__with_custom_env(self):
        dsl = instance({"DSLIBRARY_PROTO_ZZZ": "tests.test_proto_handlers.MyProtoZZZ:v=hello3"})
        r = dsl.read_resource("zzz://x")
        assert r == b'hello3'


class MyProtoZZZ(ProtocolHandler):
    def __init__(self, v="zzz"):
        self.v = v

    def open_resource(self, url: str, mode: str, **kwargs):
        return io.BytesIO(self.v.encode("utf-8"))

    def get_system_connection(self, system_type, uri: str, for_write: bool=False, **kwargs):
        class MyFS(object):
            def open(self, name):
                return io.BytesIO(b'from_fs')
        return MyFS()


"""
Loading and interpretation of model metadata.
"""
import io
import json
import os
import typing
import jsonschema
import yaml


class Metadata(object):
    """
    Information about a model, including metadata/schema about parameters, inputs and outputs.

    The mlflow MLProject file is used as a point of reference.  We extend it in various ways but aspire to remain
    compatible with it.
    """
    def __init__(self, entry_points: dict=None, uri: str=None):
        """
        :param entry_points:    Definition of each entry point.  A mapping from a name to a {} compatible with the schema.
        :param uri:             Uniquely identifies the model.
        """
        self.uri = uri or ""
        self.entry_points = {name: EntryPoint(**ep_info) for name, ep_info in (entry_points or {}).items()}

    @staticmethod
    def from_json(data):
        jsonschema.validate(data, METADATA_SCHEMA)
        return Metadata(**data)

    def to_json(self):
        return {
            "uri": self.uri,
            "entry_points": {name: info.to_json() for name, info in self.entry_points.items()}
        }

    @staticmethod
    def from_file(source: (str, io.IOBase)):
        """
        Load from a JSON or YAML file.

        :param source:   A filename or file-like object.
        """
        if isinstance(source, str):
            with open(source, 'r') as f_r:
                return Metadata.from_file(f_r)
        raw = source.read().strip()
        if not isinstance(raw, str):
            raw = raw.decode("utf-8")
        if raw.startswith("{"):
            data = json.loads(raw)
        else:
            data = yaml.safe_load(raw)
        return Metadata.from_json(data)

    @staticmethod
    def from_folder(folder: str, fill_default: bool=True):
        """
        Load metadata from specially named files in a given folder.
        """
        for f in ("metadata.yaml", "metadata.json", "MLProject"):
            fn = os.path.join(folder, f)
            if os.path.exists(fn):
                return Metadata.from_file(fn)
        # TODO check for mlflow schemas defined as YAML (where???)
        if fill_default:
            return Metadata()


class EntryPoint(object):
    """
    Describes one 'entry point' for a model, i.e. 'main', 'training', etc..
    """
    def __init__(self, command: str = None, parameters: dict = None, inputs: typing.Iterable[dict] = None,
                 outputs: typing.Iterable[dict] = None):
        self.command = command or ""
        self.parameters = {param: ParameterMetadata(**param_spec) for param, param_spec in
                           (parameters or {}).items()}
        self.inputs = {name: TableMetadata(name, **table) for name, table in (inputs or {}).items()}
        self.outputs = {name: TableMetadata(name, **table) for name, table in (outputs or {}).items()}

    def source_path(self):
        """
        Parse the source path out of the command.
        """
        # TODO code me

    @staticmethod
    def from_json(data):
        jsonschema.validate(data, METADATA_SCHEMA)
        return EntryPoint(**data)

    def to_json(self):
        return {
            "parameters": {param: param_info.to_json() for param, param_info in self.parameters.items()},
            "inputs": {name: table.to_json() for name, table in self.inputs.items()},
            "outputs": {name: table.to_json() for name, table in self.outputs.items()}
        }


class ParameterMetadata(object):
    """
    Description of parameters.
    """
    def __init__(self, type: str=None, default=None, schema: dict=None):
        self.type = type
        self.default = default
        self.schema = schema

    def to_json(self):
        out = {}
        if self.type:
            out["type"] = self.type
        if self.default is not None:
            out["default"] = self.default
        if self.schema:
            out["schema"] = self.schema
        return out

    def process_value(self, value):
        """
        Validate an input value, fill in default, and coerce to expected type.
        """
        if value is None:
            value = self.default
        if self.type == "float":
            value = float(value)
        else:
            value = str(value)
        # TODO validate 'path' and 'uri'
        if self.schema:
            value = json.loads(value)
            jsonschema.validate(value, self.schema)
        return value


class TableMetadata(object):
    """
    Description of input and output data.
    """
    def __init__(self, name: str, columns: list):
        self.name = name
        self.columns = columns

    def to_json(self):
        out = {
            "columns": self.columns
        }
        return out


METADATA_SCHEMA = {
    "comment": "Information about each distinct function the model is capable of performing.",
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "uri": {
            "comment": "Unique identifier for the model.",
            "type": "string"
        },
        # TODO add in remaining defined values for MLProject: https://www.mlflow.org/docs/latest/projects.html
        "entry_points": {
            "type": "object",
            "patternProperties": {
                ".+": {
                    "comment": "Each entry point is described here.",
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "parameters": {
                            "comment": "Parameters the model can accept for this entry point.",
                            "type": "object",
                            "patternProperties": {
                                ".+": {
                                    "comment": "Parameter is described here.",
                                    "oneOf": [
                                        {
                                            "comment": "MLProject files let you specify a very constrained data type",
                                            "type": "string", "enum": ["path", "uri", "float", "string"]
                                        },
                                        {
                                            "comment": "MLProject files let you define a parameter with these fields.",
                                            "type": "object",
                                            "required": ["type"],
                                            "additionalProperties": False,
                                            "properties": {
                                                # NOTE: the 'path' and 'uri' types imply a file-like input data source
                                                "type": {"type": "string", "enum": ["path", "uri", "float", "string"]},
                                                "default": {"comment": "Default value"},
                                                "schema": {
                                                    # EXTENSION to MLProject to support validation of structured data types
                                                    "comment": "For type=string, a JSON schema can be imposed.",
                                                    "type": "object",
                                                    "properties": {
                                                        "type": {},
                                                        "required": {"type": "array"},
                                                        "properties": {},
                                                        "additionalProperties": {"type": "boolean"},
                                                        "pattern": {"type": "string"}
                                                        # additional JSON schema fields are allowed here
                                                    }
                                                }
                                            }
                                        }
                                    ]
                                }
                            }
                        },
                        "command": {
                            "comment": "Command to execute for this model",
                            "type": "string"
                        },
                        # EXTENSION to MLProject to define schemas for input data sources
                        "inputs": {
                            "comment": "Information about data inputs",
                            "type": "object",
                            "patternProperties": {
                                ".+": {
                                    "$ref": "#/definitions/table"
                                }
                            }
                        },
                        # EXTENSION to MLProject to define schemas for data outputs
                        "outputs": {
                            "comment": "Information about data inputs",
                            "type": "object",
                            "patternProperties": {
                                ".+": {
                                    "$ref": "#/definitions/table"
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    "definitions": {
        "table": {
            "type": "object",
            "properties": {
                "columns": {
                    "comment": "Column names and data types, for columnar data.",
                    "type": "array",
                    "items": {
                        "type": "object",
                        "additionalProperties": False,
                        "required": ["name"],
                        "properties": {
                            "name": {
                                "comment": "Short name by which the data will be referenced from the code.",
                                "type": "string"
                            },
                            "type": {
                                "comment": "Broad data type",
                                "enum": ["number", "string", "integer", "boolean"]
                            },
                            "required": {
                                "comment": "Which fields are required.",
                                "type": "array"
                            }
                        }
                    }
                }
            }
        }
    }
}



import unittest
import mock
import io
import os
import numpy
import pandas
import shutil
import tempfile
import yaml
import dask.dataframe

from dslibrary import DSLibrary
from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS, DSLibraryException, DSLibraryDataFormatException
from dslibrary.metadata import Metadata
from dslibrary.transport.to_local import DSLibraryLocal


class TestFront(unittest.TestCase):

    def test_metadata(self):
        """
        The base class doesn't know how to get metadata so it returns a null instance.
        """
        dsl = DSLibrary()
        m = dsl.get_metadata()
        assert isinstance(m, Metadata)
        assert m.uri == ""
        assert m.entry_points == {}

    def test_open_resource__input_mapping(self):
        """
        Options for input come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"inputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", option1=2, option2=22)
        assert log[0] == ('aaa', 'rb', {'option1': 1, 'option2': 22, 'option3': 3}), log[0]

    def test_open_resource__output_mapping(self):
        """
        Options for output come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"outputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", 'w', option1=2, option2=22)
        self.assertEqual(log[0], ('aaa', 'w', {'option1': 1, 'option2': 22, 'option3': 3}))

    def test_open_resource__bypass_mapping(self):
        """
        Most inputs and outputs are mapped, but you can bypass mapping and specify a URI to open or a local file.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_resource("s3://bucket/path", open_option_1=1)
        assert log[0] == ('s3://bucket/path', 'rb', {'open_option_1': 1})
        MyInst().open_resource("./local/file")
        assert log[1] == ('./local/file', 'rb', {})

    def test_open_model_binary(self):
        """
        The default implementation just assumes certain filenames for the 'model-binary' data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_model_binary()
        assert log[0] == ('model-binary', 'rb', {})
        MyInst().open_model_binary("part1")
        assert log[1] == ('model-binary/part1', 'rb', {})

    def test_set_evaluation_result(self):
        """
        Model evaluation code can use this hook to report whether the model passed or failed.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        MyInst().set_evaluation_result(True)
        MyInst().set_evaluation_result(False, reason="because")
        assert log[0][0] == log[1][0] == EVALUATION_RESULT_ALIAS
        r = log[0][1].getvalue()
        assert r == '{"uri": "", "success": true}\n', r
        r = log[1][1].getvalue()
        assert r == '{"uri": "", "success": false, "reason": "because"}\n', r

    def test_get_sql_connection__mapping(self):
        """
        The normal use case is have the caller supply all the connection information.
        """
        dsl = DSLibrary(spec={"inputs": {"db": {"uri": "mysql://host/db", "username": "u"}}})
        def connect(user, host, database, **kwargs):
            assert user == "u"
            assert host == "host"
            assert database == "db"
            return "CONN"
        with mock.patch("pymysql.connect", connect):
            r = dsl.get_sql_connection("db")
            assert r == "CONN"

    def test_write_resources__dataframes_and_series(self):
        """
        Several types are supported for columnar data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        dsl = MyInst()
        # series
        dsl.write_resource("x", pandas.Series([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'col1\n1\n2\n3\n')
        log.clear()
        # numpy array
        dsl.write_resource("x", numpy.array([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'col1\n1\n2\n3\n')
        log.clear()
        # dataframe
        dsl.write_resource("x", pandas.DataFrame({"y": [1, 2, 3]}))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'y\n1\n2\n3\n')
        log.clear()

    def test_default_metrics_output(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project)
        dsl.log_metric("x", 1)
        r = dsl.get_last_metric("x")
        assert r.value == 1
        # verify it was written as JSON
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("{")
        shutil.rmtree(project)

    def test_alt_metrics_output1(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"format": "csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_alt_metrics_output2(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, "metrics.csv")
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_model_pickling(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        # save and restore
        my_model = {"x": 1}
        dsl.save_pickled_model(my_model)
        restored = dsl.load_pickled_model()
        assert restored == my_model
        # verify local storage
        assert os.path.exists(project + "/model-binary")
        shutil.rmtree(project)

    def test_open_resource__default_uri(self):
        """
        Specify a name and a default URI, for cleaner overriding.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, kwargs))
                return "H"
        MyInst().open_resource("x", uri="default")
        MyInst(spec={"inputs": {"x": {"uri": "override", "option1": 1}}}).open_resource("x", uri="default")
        self.assertEqual(log, [('default', {}), ('override', {'option1': 1})])

    def test_load_dataframe__default_uri(self):
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append(path)
                return io.BytesIO(b"x\n1\n2")
        df = MyInst().load_dataframe("x", uri="default.csv")
        df = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).load_dataframe("x", uri="default.csv")
        self.assertEqual(log[0], 'default.csv')
        self.assertEqual(log[-1], 'override.csv')

    def test_load_dataframe__from_sql(self):
        log = []
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        log.append(sql)
                    def __iter__(self):
                        return iter([(1,), (2,)])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        log.append("close")
                return DbConn()
        df = MyInst().load_dataframe("x", sql_table="table1")
        self.assertEqual(list(df.x), [1, 2])
        self.assertEqual(log, ['SELECT * from table1', 'close'])

    def test_load_dataframe__from_nosql(self):
        class MyInst(DSLibrary):
            def get_nosql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class MyNoSql(object):
                    def query(self, collection, **kwargs):
                        assert collection == "table1"
                        return [{"x": 1}, {"x": 2}]
                return MyNoSql()
        df = MyInst().load_dataframe("x", nosql_collection="table1")
        self.assertEqual(list(df.x), [1, 2])

    def test_load_dataframe__sql__custom_open_args(self):
        """
        Custom arguments can be passed through to get_sql_connection().
        """
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, database: str=None, for_write: bool=False, **kwargs):
                assert resource_name == "engine"
                assert database == "db"
                assert kwargs["custom1"] == 123
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        pass
                    def __iter__(self):
                        return iter([])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        pass
                return DbConn()
        MyInst().load_dataframe("engine", sql_table="table1", database="db", custom1=123)

    def test_open_resource__named_filesystem(self):
        """
        A resource can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                assert kwargs.get("custom1") == 111
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "r"
                        return "FH"
                return FS()
        r = MyInst().open_resource("path", mode="r", filesystem="fs", custom1=111)
        assert r == "FH"

    def test_load_dataframe__named_filesystem(self):
        """
        A dataframe can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "rb"
                        return io.BytesIO(b"x\n1\n2\n3")
                return FS()
        df = MyInst().load_dataframe("path", filesystem="fs", format="csv")
        assert list(df.x) == [1, 2, 3]

    def test_rw_run_data(self):
        writes = {}
        class MyInst(DSLibrary):
            def open_run_data(self, filename: str, mode: str='rb'):
                if mode == 'rb':
                    v = writes[filename].getvalue()
                    return io.BytesIO(v)
                wr = io.BytesIO()
                writes[filename] = wr
                wr.close = lambda: None
                return wr
        dsl = MyInst()
        df = pandas.DataFrame({"x": [1, 2]})
        dsl.write_run_data("x.csv", df)
        self.assertEqual(writes["x.csv"].getvalue(), b'x\n1\n2\n')
        df2 = dsl.load_dataframe("x.csv", run_data=True)
        self.assertEqual(list(df2.x), list(df.x))
        self.assertEqual(dsl.read_run_data("x.csv"), b'x\n1\n2\n')

    def test_open_args_passed_through_to_opener(self):
        log = []
        inst = self
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append(path)
                inst.assertEqual(kwargs, {'arg1': 123})
                return io.BytesIO(b"abc")
        # just a filename
        fh = MyInst().open_resource("x", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # uri overrides name
        fh = MyInst().open_resource("x", uri="default.csv", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # input mapping maps 'x' to something else
        fh = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).open_resource("x", uri="default.csv", arg1=123, format="xyz")
        assert fh.read() == b"abc"
        # verify filenames
        self.assertEqual(log, ['x', 'default.csv', 'override.csv'])

    def test_setup_code_paths(self):
        """
        You can cause 'sys.path' to be extended to include particular folders.
        This is normally done using an environment variable.  Here we are just testing the method that applies the
        changes.
        """
        inst = DSLibrary()
        inst._spec = {"code_paths": ["a", "b"]}
        mock_paths = []
        with mock.patch("sys.path", mock_paths):
            inst._setup_code_paths()
        self.assertEqual(mock_paths, ["a", "b"])

    def test_abstract_base_methods(self):
        """
        The methods that have to be filled in raise exceptions when called.
        """
        inst = DSLibrary()
        self.assertRaises(DSLibraryException, lambda: inst._opener("path", mode='r'))
        self.assertRaises(DSLibraryException, lambda: inst.open_run_data("fn"))

    def test_get_filesystem_connection(self):
        """
        The default filesystem accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_filesystem", lambda **k: log.append(k) or "FS"):
            fs = inst.get_filesystem_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "FS")
        self.assertEqual(log, [{'uri': 'resource', 'for_write': True, 'arg1': 123}])

    def test_get_sql_connection(self):
        """
        The default SQL accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_database", lambda **k: log.append(k) or "NoSqlConn"):
            fs = inst.get_sql_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "NoSqlConn")
        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])

    def test_get_nosql_connection(self):
        """
        The default NoSQL accessor.
        """
        inst = DSLibrary()
        log = []
        with mock.patch("dslibrary.front.connect_to_nosql", lambda **k: log.append(k) or "SqlConn"):
            fs = inst.get_nosql_connection("resource", for_write=True, arg1=123)
            self.assertEqual(fs, "SqlConn")
        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])

    def test_log_mlflow(self):
        inst = DSLibrary()
        inst._mlflow_all = True
        log = []
        with mock.patch("mlflow.log_param", lambda k, v: log.append((k, v))):
            inst.log_param("x", 1)
        self.assertEqual(log, [("x", 1)])
        log.clear()
        with mock.patch("mlflow.log_metric", lambda *a: log.append(a)):
            inst.log_metric("q", 222)
            inst.log_metrics({"z": 9}, step=2)
        self.assertEqual(log, [('q', 222, 0), ('z', 9, 2)])
        log.clear()
        with mock.patch("mlflow.log_dict", lambda *a: log.append(a)):
            inst.log_dict({"x": 1}, "f.json")
        self.assertEqual(log, [({'x': 1}, 'f.json')])
        log.clear()
        with mock.patch("mlflow.log_artifact", lambda *a: log.append(a)):
            inst.log_artifact("f_local", "f_store")
        self.assertEqual(log, [('f_local', 'f_store')])
        log.clear()
        # not a log function but close enough
        self.assertRaises(DSLibraryException, lambda: inst.get_metrics("x"))

    def test_mlflow_start_end(self):
        inst = DSLibrary()
        inst._mlflow_all = True
        log = []
        with mock.patch("mlflow.mlflow.start_run", lambda: log.append("start")):
            with mock.patch("mlflow.mlflow.end_run", lambda: log.append("end")):
                with mock.patch("mlflow.active_run", lambda: log.append("active_run")):
                    # TODO mock this: mlflow.active_run().info.run_id
                    with inst.start_run():
                        pass
        self.assertEqual(log, ['start', 'active_run', 'end'])

    def test_yaml(self):
        data = [
            {"age": 41, "name": {"first": "John", "last": "Smith"}}
        ]
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.StringIO(yaml.dump(data))
        df = MyInst().load_dataframe("data.yaml")
        assert list(df.age) == [41]
        assert list(df["name.first"]) == ["John"]

    def test_xlsx(self):
        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return open(path, mode)
        df = MyInst().load_dataframe(fn)
        self.assertEqual(list(df.a), [2, 7])
        self.assertEqual(list(df.b), [5, 9])

    def test_xlsx__unseekable_stream(self):
        """
        The xlsx parser requires a seekable stream, so when we supply an unseekable stream
        it does some special processing.
        """
        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                with open(path, mode) as f_r:
                    data = f_r.read()
                class Stream(io.RawIOBase):
                    def __init__(self, data):
                        self.data = data
                    def read(self, n=None):
                        if n is None:
                            n = len(self.data)
                        out = self.data[:n]
                        self.data = self.data[n:]
                        return out
                    def readable(self):
                        return True
                    def seekable(self):
                        return False
                    def __enter__(self):
                        return self
                    def __exit__(self, exc_type, exc_val, exc_tb):
                        pass
                return Stream(data)
        df = MyInst().load_dataframe(fn)
        self.assertEqual(list(df.a), [2, 7])
        self.assertEqual(list(df.b), [5, 9])

    def test_load_dataframe__strip_open_args(self):
        inst = self
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                inst.assertEqual(path, "x")
                inst.assertEqual(kwargs, {})
                return io.StringIO("x\n1\n2")
        r = MyInst().load_dataframe("x", dask=False, format="csv", format_options={}, fallback_to_text=False)
        assert list(r.x) == [1, 2]

    def test_load_dataframe__empty_cell_case(self):
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.StringIO("x0,x1\nA,\n,B\nD,C")
        df = MyInst().load_dataframe("x.csv")
        self.assertEqual(list(df.fillna(value="").x0), ["A", "", "D"])

    def test_json_case(self):
        x = '{"uri":"","run_id":"","user":"","time":1634234451.674405098,"name":"x","value":1,"step":0}'
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.StringIO(x)
        df = MyInst().load_dataframe("x")
        assert "run_id" in df.columns
        assert "name" in df.columns

    def test_write_resource_sql_table(self):
        log = []
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, for_write: bool=False, database=None, **kwargs):
                assert resource_name == "my_pg"
                assert database == "db4"
                class DbCursor(object):
                    def execute(self, sql, params):
                        log.append((sql, params))
                    def __iter__(self):
                        return iter([])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        pass
                return DbConn()
        # JSON data
        MyInst().write_resource("my_pg", {"x": [1, 2, 3]}, append=False, sql_table="tbl1", database="db4")
        self.assertEqual(log, [
            ('DROP TABLE IF EXISTS tbl1', []),
            ('CREATE TABLE IF NOT EXISTS tbl1 (x INTEGER)', []),
            ('INSERT INTO tbl1 (x) VALUES (%s), (%s), (%s)', [1, 2, 3])
        ])
        log.clear()
        # a series
        MyInst().write_resource("my_pg", pandas.Series([1, 2, 3]), append=False, sql_table="tbl1", database="db4")
        self.assertEqual(log, [
            ('DROP TABLE IF EXISTS tbl1', []),
            ('CREATE TABLE IF NOT EXISTS tbl1 (col1 INTEGER)', []),
            ('INSERT INTO tbl1 (col1) VALUES (%s), (%s), (%s)', [1, 2, 3])
        ])

    def test_load_dataframe__dask_autodetect(self):
        """
        Size-based threshold.
        """
        with tempfile.NamedTemporaryFile(suffix=".csv") as f_tmp:
            f_tmp.write(b"x\n1\n2")
            f_tmp.flush()
            class MyInst(DSLibrary):
                def _opener(self, path: str, mode: str, **kwargs):
                    return io.BytesIO(b"x\n1\n2")
            # specifically don't use dask
            df = MyInst().load_dataframe(f_tmp.name, dask=False)
            self.assertIsInstance(df, pandas.DataFrame)
            # specifically do use dask
            df = MyInst().load_dataframe(f_tmp.name, dask=True)
            assert hasattr(df, "dask")
            # high threshold, don't use dask
            df = MyInst().load_dataframe(f_tmp.name, dask=1000)
            self.assertIsInstance(df, pandas.DataFrame)
            # low threshold, do use dask
            df = MyInst().load_dataframe(f_tmp.name, dask=3)
            assert hasattr(df, "dask")

    def test_load_dataframe__from_stream(self):
        """
        An already opened stream can be supplied instead of a named entity.
        """
        stream = io.BytesIO(b"x\n1\n2\n12")
        df = DSLibrary().load_dataframe(stream)
        self.assertEqual(list(df.x), [1, 2, 12])
        stream = io.StringIO("x\n1\n2\n12")
        df = DSLibrary().load_dataframe(stream)
        self.assertEqual(list(df.x), [1, 2, 12])

    def test_load_dataframe__from_iterable(self):
        """
        Data can be supplied to turn into a dataframe.
        """
        df = DSLibrary().load_dataframe({"x": [11, 12, 13]})
        self.assertEqual(list(df.x), [11, 12, 13])
        df = DSLibrary().load_dataframe([{"x": 9}, {"x": 8}])
        self.assertEqual(list(df.x), [9, 8])

    def test_hash_in_uri_for_formmatting_argumets(self):
        """
        Since the '#' has no effect on information returned from a URL, it is a safe place to place formatting arguments.
        """
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                assert path == "x", "the '#' is removed before opening the file"
                return io.BytesIO(b"a,b\n1,2\n2,3")
        # here the default format sniffer guesses there are two columns
        df = MyInst().load_dataframe("x")
        self.assertEqual(list(df.columns), ["a", "b"])
        # here we force the delimited to be a linefeed and now there is only one column
        df = MyInst().load_dataframe("x#delimiter=%0a")
        self.assertEqual(list(df.columns), ["a,b"])

    def test_hdf(self):
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return open(path, mode)
        fn = os.path.join(os.path.dirname(__file__), "test_data/my.hdf")
        df = MyInst().load_dataframe(fn)
        assert list(df.columns) == ["x"]
        assert list(df.x) == [1, 2, 3]

    def test_report_encoding_error(self):
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.BytesIO(b'a\n1\n2\n3\n\xC0\xC0\xFF\n4\n5\n6')
        try:
            MyInst().load_dataframe("x.csv")
            self.fail("expected error")
        except DSLibraryDataFormatException as err:
            assert "invalid start byte" in err.message
            assert "offset=8" in err.message
        df = MyInst().load_dataframe("x.csv", format_options={"encoding": 'iso8859-1'})
        self.assertEqual(list(df.columns), ["a"])

    def test_report_json_format_error(self):
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                return io.BytesIO(b'{"x":' + b' '*1000)
        try:
            MyInst().load_dataframe("x.json")
            self.fail("expected error")
        except DSLibraryDataFormatException as err:
            assert "JSON format error" in err.message

    def test_rw_folder(self):
        """
        Dask writes data into folders, in the manner of spark.
        """
        for dataframe_type in ("p", "d"):
            tmpdir = tempfile.mkdtemp()
            try:
                class MyInst(DSLibrary):
                    def _opener(self, path: str, mode: str, **kwargs):
                        return open(path, mode)
                fn = os.path.join(tmpdir, "x.csv")
                df = pandas.DataFrame({"x": list(range(100))})
                if dataframe_type == "d":
                    df = dask.dataframe.from_pandas(df, npartitions=2)
                parts = MyInst().write_resource(fn, df, dask=True)
                self.assertEqual(os.listdir(tmpdir), ["x.csv"])
                if dataframe_type == "d":
                    self.assertEqual(os.listdir(fn), ["0.part", "1.part"])
                    assert len(parts) == 2
                else:
                    self.assertEqual(os.listdir(fn), ["0.part"])
                    assert len(parts) == 1
                df = MyInst().load_dataframe(fn, dask=True, format_options={"delimiter": "\n"}).compute()
                self.assertEqual(list(df.x), list(range(100)))
            finally:
                shutil.rmtree(tmpdir)

    def test_write_dask_csv(self):
        """
        Dask is able to write to a single CSV file.
        """
        files = {}
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                files[path] = fh = io.BytesIO()
                fh.close = lambda: None
                return fh
        df = pandas.DataFrame({"x": list(range(7))})
        MyInst().write_resource("x.csv", dask.dataframe.from_pandas(df, npartitions=2), dask=False)
        self.assertEqual(files["x.csv"].getvalue(), b"x\n0\n1\n2\n3\n4\n5\n6\n")




import unittest
import pickle
import tempfile
import zipfile
import io

from dslibrary.front import DSLibraryException
from dslibrary.transport.to_mmlibrary import DSLibraryViaMMLibrary


class TestToMMLibrary(unittest.TestCase):

    def test_resources(self):
        resources = {}
        class MyMM(object):
            def get_binary_from_resource(self, resource_name):
                return resources[resource_name]
            def save_binary_to_resource(self, resource_name, data):
                resources[resource_name] = data
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        with dsl.open_resource("a", mode='w') as f_w:
            f_w.write("hello")
        with dsl.open_resource("a", mode="r") as f_r:
            assert f_r.read() == "hello"
        with dsl.open_resource("a", mode='wb') as f_w:
            f_w.write(b"hello")
        with dsl.open_resource("a", mode="rb") as f_r:
            assert f_r.read() == b"hello"

    def test_parameters(self):
        params = {"x": 1}
        class MyMM(object):
            def get_argument(self, name):
                return params[name]
            param_dictionary = params
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        r = dsl.get_parameters()
        assert r == {"x": 1}
        dsl.get_parameter("x")
        assert dsl.get_parameter("y", 222) == 222

    def test_open_run_data(self):
        class MyMM(object):
            def __init__(self):
                self.data = b''
            def save_temporary_data(self, data):
                self.data = data
            def get_temporary_data(self):
                return self.data
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        with dsl.open_run_data("x", mode='w') as f_w:
            f_w.write("this is x")
        with dsl.open_run_data("y", mode='wb') as f_w:
            f_w.write(b"this is y")
        with dsl.open_run_data("x", mode='r') as f_r:
            assert f_r.read() == "this is x"
        with dsl.open_run_data("y", mode='rb') as f_r:
            assert f_r.read() == b"this is y"
        self.assertRaises(FileNotFoundError, lambda: dsl.open_run_data("z", mode='rb'))

    def test_load_dataframe__default_options(self):
        """
        You can supply default options for loading a CSV file.
        """
        class MyMM(object):
            def get_binary_from_resource(self, resource_name):
                return b'a\tb\n1\t2\n3\t4\n'
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        df = dsl.load_dataframe("x.csv", format_options={"delimiter": "\t"})
        assert list(df.columns) == ["a", "b"]
        assert list(df.a) == [1, 3]

    def test_get_sql_connection(self):
        class MyMM(object):
            def get_db_connection(self, resource_name):
                return resource_name
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        r = dsl.get_sql_connection("x")
        self.assertEqual(r, "x")

    def test_load_pickled_model(self):
        with tempfile.NamedTemporaryFile() as f_tmp:
            pickle.dump({"x": 1}, f_tmp)
            f_tmp.flush()
            class MyMM(object):
                def get_model(self):
                    return f_tmp.name
            dsl = DSLibraryViaMMLibrary(_mm=MyMM())
            r = dsl.load_pickled_model()
            self.assertEqual(r, {"x": 1})

    def test_save_pickled_model(self):
        log = []
        class MyMM(object):
            def new_version(self, data):
                log.append(data)
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        dsl.save_pickled_model({"x": 1})
        self.assertEqual(pickle.loads(log[0]), {"x": 1})

    def test_open_model_binary__parts(self):
        """
        You can send a ZIP file for the model-binary, and then use the 'part' argument of open_model_binary() to access them.
        """
        with tempfile.NamedTemporaryFile() as f_tmp:
            with zipfile.ZipFile(f_tmp, mode='w') as zf:
                zf.writestr("a", "AAA")
                zf.writestr("b", "BBB")
            f_tmp.flush()
            class MyMM(object):
                def get_model(self):
                    return f_tmp.name
            dsl = DSLibraryViaMMLibrary(_mm=MyMM())
            with dsl.open_model_binary("a", "r") as f_r:
                self.assertEqual(f_r.read(), "AAA")
            with dsl.open_model_binary("b", "r") as f_r:
                self.assertEqual(f_r.read(), "BBB")

    def test_named_filesystem_via_get_db_connection(self):
        """
        Access to something like an S3 bucket - with no implementation of a bucket_name override.
        """
        class MyMM(object):
            def get_db_connection(self, resource_name):
                assert resource_name == "fs123"
                class FH(object):
                    def open(self, filename, mode, **kwargs):
                        assert filename == "name"
                        assert mode == "r"
                        return io.StringIO("xyz")
                return FH()
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        # bucket_name can't be overridden in this case
        self.assertRaises(DSLibraryException, lambda: dsl.open_resource("name", mode="r", filesystem="fs123", bucket_name="b1"))
        # we can use the data connection without a bucket_name override
        r = dsl.open_resource("name", mode="r", filesystem="fs123")
        assert r.read() == "xyz"

    def test_get_db_connection__db_or_bucket(self):
        """
        Read from s3 with a bucket_name override in the mmlibrary implementation.
        """
        class MyMM(object):
            def get_db_connection(self, resource_name, _db_or_bucket=None):
                assert resource_name == "fs123"
                assert _db_or_bucket == "b1"
                class FH(object):
                    def open(self, filename, mode, **kwargs):
                        assert filename == "name"
                        assert mode == "r"
                        return io.StringIO("xyz")
                return FH()
        dsl = DSLibraryViaMMLibrary(_mm=MyMM())
        r = dsl.open_resource("name", mode="r", filesystem="fs123", bucket_name="b1")
        assert r.read() == "xyz"



import unittest
import random
import os
import pandas
from pandas.core.dtypes.inference import is_file_like

from dslibrary.utils.filechunker import ChunkedFileReader, ChunkedFileWriter


class TestFileChunkers(unittest.TestCase):

    def test_rd(self):
        data = "1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert f.name == "x"
        assert f.read(3) == "123"
        assert f.read() == "4567890"
        f.seek(1)
        assert f.read(4) == "2345"
        f.close()

    def test_rd_b(self):
        data = b"1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='rb')
        assert f.name == "x"
        assert f.read(3) == b"123"
        assert f.read() == b"4567890"
        f.seek(1)
        assert f.read(4) == b"2345"
        f.close()

    def test_rd_minChunk(self):
        data = "1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        f._min_chunk = 4
        assert f.read(3) == "123"
        assert f._chunk == "1234"
        assert f._offset == 3
        assert f._pos == 3
        assert f.read(2) == "45"
        assert f._chunk == "5678"
        assert f._offset == 1
        assert f._pos == 5
        f.seek(6)
        assert f._offset == 2
        assert f._pos == 6
        assert f.read(2) == "78"

    def test_seek_misc(self):
        data = "1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert f.read(3) == "123"
        assert f.tell() == 3
        f.seek(2)
        assert f.tell() == 2
        assert f.read(3) == "345"
        assert f.tell() == 5
        f.seek(1, 1)
        assert f.tell() == 6
        assert f.read() == "7890"
        f.seek(-2, 2)
        assert f.tell() == 8
        assert f.read() == "90"
        assert f.tell() == 10

    def test_seek_from_end(self):
        data = "1234567890"
        for mode in ["constant", "function", "unknown"]:
            f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r', size=10 if mode == "constant" else lambda: 10 if mode == "function" else None)
            if mode == "unknown":
                self.assertRaises(Exception, lambda: f.seek(0, 2))
                continue
            f.seek(0, 2)
            assert f.tell() == 10
            f.seek(0, 0)
            f.seek(-2, 2)
            assert f.tell() == 8
            assert f.read() == "90"
            assert f.tell() == 10

    def test_rd_max_read(self):
        data = "x"*200
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        f._max_read = 100
        assert f.read() == "x"*100
        assert f.read() == "x"*100
        assert f.read() == ""

    def test_rd_direct(self):
        data = "x"*100
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert f.read() == "x"*100

    def test_rd_fix_type(self):
        data = "1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='rb')
        assert f.read(3) == b"123"
        data = b"1234567890"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert f.read(3) == "123"

    def test_rd_lines(self):
        data = "a\nb\nc\nd"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        rr = list(f)
        assert rr == ["a", "b", "c", "d"]

    def test_wr_misc(self):
        out = ['x']
        def wr(data, append):
            if not append:
                out[0] = ""
            out[0] += data
        f = ChunkedFileWriter("name", wr, mode='w')
        assert f.name == "name"
        f.write("a")
        f.write("b")
        f.flush()
        f.close()
        assert out == ["ab"]

    def test_wr_append(self):
        out = ['x']
        def wr(data, append):
            if not append:
                out[0] = ""
            out[0] += data
        f = ChunkedFileWriter("name", wr, mode='wa')
        assert f.name == "name"
        f.write("a")
        f.write("b")
        f.close()
        assert out == ["xab"]

    def test_wr_mpu(self):
        """
        Write with multi-part-upload hints.
        """
        log = []
        def wr(data, append, hint):
            log.append((data, append, hint))
            if hint.startswith("start"):
                return "UUU"
        f = ChunkedFileWriter("name", wr, mode='w', chunk_size=100)
        assert f.name == "name"
        f.write("a"*80)
        f.write("b"*80)
        f.close()
        assert log[0] == ("a"*80 + "b"*20, False, "start:-1")
        assert log[1] == ("b"*60, True, "end:UUU:1")

    def test_wr_mpu_1(self):
        """
        Write with multi-part-upload hints.
        """
        log = []
        def wr(data, append, hint=None):
            log.append((data, append, hint))
            if hint == "start":
                return "UUU"
        f = ChunkedFileWriter("name", wr, mode='w', chunk_size=100)
        assert f.name == "name"
        f.write("a"*80)
        f.close()
        assert log[0] == ("a"*80, False, None)

    def test_random_lines(self):
        lengths = [random.randint(0, 40) for _ in range(1000)]
        out = []
        with ChunkedFileWriter("", lambda data, append: out.append(data), chunk_size=101) as wr:
            for n, ll in enumerate(lengths):
                ch = chr(ord("0") + (n % 9))
                wr.write(ch*ll + "\n")
        raw = "".join(out)
        rd = ChunkedFileReader("", lambda a, b: raw[a:b], chunk_size=100)
        l_out = [len(line) for line in rd]
        assert l_out == lengths

    def test_post_slice(self):
        data = "a\nb\nc\nd"
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert list(f[:2]) == ["a", "b"]
        f.seek(0)
        assert list(f[2:]) == ["c", "d"]
        f.seek(0)
        assert list(f[::2]) == ["a", "c"]
        self.assertRaises(Exception, lambda: list(f[-2:]))
        self.assertRaises(Exception, lambda: list(f[0:-2]))
        self.assertRaises(Exception, lambda: list(f[0:1:-1]))

    def test_eof_issue(self):
        data = 'x' * 100
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert len(f.read(1000)) == 100
        d = f.read(1000)
        assert not d and f._eof
        d = f.read(1000)
        assert not d and f._eof

    def test_pandas_ok(self):
        data = 'x' * 100
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='r')
        assert is_file_like(f)

    def test_pandas_ok2(self):
        data = 'x,y\n1,10\n2,20'
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='rb')
        df = pandas.read_csv(f)
        assert list(df.columns) == ["x", "y"]

    def test_pandas_ok3(self):
        fn = os.path.join(os.path.dirname(__file__), "../test_data/two_cols.xlsx")
        with open(fn, 'rb') as f:
            data = f.read()
        f = ChunkedFileReader("x", lambda a, b: data[a:b], mode='rb', size=len(data))
        df = pandas.read_excel(f, engine="openpyxl")
        assert list(df.columns) == ["a", "b"]

    def test_capabilities(self):
        f = ChunkedFileReader("x", lambda a, b: "", mode='r')
        assert f.readable() is True
        assert f.writable() is False
        assert f.seekable() is True
        def wr(data, append):
            pass
        f = ChunkedFileWriter("name", wr, mode='w')
        assert f.writable() is True
        assert f.readable() is False
        assert f.seekable() is False



"""
A model management engine that wraps MLFlow.  We have the option of using MLFlow only for tracking, or for
both tracking and execution.

When using for tracking only, the model communicates with the engine over REST and the engine calls MLFlow on its
behalf.  When using for both tracking and execution, the model calls mmlibrary, which delegates to MLFlow for some
tasks (like reporting metrics), but calls back to the engine for others, like accessing input data.
"""
import os
import tempfile

from mmlibrary.front import MMLibrary, Metric
from mmlibrary.metadata import Metadata

try:
    import mlflow
except ImportError:
    mlflow = None

IMAGE_EXTENSIONS = ("png", "jpg", "jpeg", "gif")


class RunInMLFlow(object):
    def run(self):
        pass


class MMLibraryToMLFlow(MMLibrary):
    """
    The executing model calls the main MMLibrary entry points, which transfer control, one way or another,
    to this class.

    Here, we call MLFlow methods on the running model's behalf.

    WILL MLFLOW BE EXECUTING THE MODELS?  Not necessarily.
      - if it is, then hopefully it doesn't mind the tracking calls coming from an unexpected host
      - if it isn't, then we will need to delegate the run() calls somewhere

    If supplied, input data requests are delegated to data_handler.
    """
    def __init__(self, data_handler=None, execution_handler=None):
        if not mlflow:
            raise Exception("The 'mlflow' package is required to use this interface")
        self._data_handler = data_handler
        self._execution_handler = execution_handler
        self._run = None
        self._experiment_id = self._run.info.experiment_id

    def get_metadata(self, other_model_uri: str=None) -> Metadata:
        # TODO URI & metadata needed for model being executed
        return Metadata()

    def start(self):
        self._run = mlflow.active_run()
        return mlflow.start_run()

    def end(self):
        return mlflow.end_run()

    def open_resource(self, resource_name: str, mode: str = 'r', context_uri: str=None):
        if "r" in mode:
            if self._data_handler:
                return self._data_handler.open_resource(resource_name, mode, context_uri)
            # TODO error -
        else:
            # TODO capture the content, then on close() call write_resource()
            pass

    def write_resource(self, filename: str, content=None, append: bool=False, target_path: str=None):
        """
        All the output is redirected to mlflow.
        :param filename:
        :param content:
        :param append:
        :param target_path:
        :return:
        """
        target_filename = self._target_filename(filename, target_path)
        file_extn = os.path.splitext(filename)[1].strip(".").lower()
        target_extn = os.path.splitext(target_path or "")[1].strip(".").lower()
        if isinstance(content, dict):
            return mlflow.log_dict(content, target_filename)
        if isinstance(content, str):
            return mlflow.log_text(content, target_filename)
        if file_extn in IMAGE_EXTENSIONS or target_extn in IMAGE_EXTENSIONS:
            if content is None:
                with open(filename, 'rb') as f_r:
                    content = f_r.read()
            return mlflow.log_image(content, target_filename)
        if hasattr(content, "savefig"):
            return mlflow.log_figure(content, target_filename)
        if content is not None:
            with tempfile.NamedTemporaryFile(mode='w' if isinstance(content, str) else 'wb') as f_tmp:
                f_tmp.write(content)
            mlflow.log_artifact(f_tmp.name, filename or target_path)
        else:
            mlflow.log_artifact(filename, target_path)

    def save_metric(self, metric_name: str, metric_value):
        # TODO mlflow adds 'step', which disambiguates among metrics with the same name from the same run
        mlflow.log_metric(metric_name, metric_value)

    def save_metrics(self, metrics: dict) -> None:
        mlflow.log_metrics(metrics)

    def get_metrics(self, metric_name: str = None, uri: str = None, time_range: (list, tuple) = None, limit: int = None):
        if uri or time_range or limit:
            # TODO finish implementing this
            pass
        run = mlflow.get_run(self._run.info.run_id)
        metrics = run.data.metrics
        t0 = run.start_time
        for k, v in metrics.items():
            yield Metric(run_id=self._run.info.run_id, uri=self._experiment_id, time=t0, name=k, value=v)

    def get_db_connection(self, connection_name: str, database: str = None):
        if self._data_handler:
            # TODO returning a DBI interface over REST won't do.  I think we need a more clever means of delegation here
            return self._data_handler.get_sql_connection(connection_name, database)
        # TODO error, no data handler

    def run(self, code_uri: str, arguments: dict = None, callback=None):
        if self._execution_handler:
            # TODO the callback won't work over REST, of course...
            return self._execution_handler.run(code_uri, arguments, callback)
        # TODO translate call into MLFlow-ese...
        #  mlflow.run(uri, entry_point='main', version=None, parameters=None, docker_args=None, experiment_name=None, experiment_id=None, backend='local', backend_config=None, use_conda=True, storage_dir=None, synchronous=True, run_id=None)



"""
Increment the version number and deploy to PyPi.
"""
import os
import glob
import re
import sys
import subprocess

try:
    import twine
except ImportError:
    print("Please install twine first:   pip install twine")
    raise


def incr_version(v: str):
    v0 = tuple(int(part) for part in v.split("."))
    v1 = v0[:-1] + (v0[-1] + 1,)
    return ".".join(str(_) for _ in v1)


def replace_version(code: str, match, new_v: str):
    v_start, v_end = match.regs[1]
    return code[:v_start] + new_v + code[v_end:]


# parse options
increment = True
write = True
for arg in sys.argv:
    if arg == "--no-increment":
        increment = False
    if arg == "--no-write":
        write = False

# determine version from setup.py
fn1 = "setup.py"
fn2 = "dslibrary/__init__.py"
ptn_version1 = re.compile(r'\n\s*version\s*=\s*"([\d.]+)",\s*\n')
ptn_version2 = re.compile(r'\n__version__\s*=\s*"([\d.]+)"\s*\n')
with open(fn1, "r") as f_r:
    code1 = f_r.read()
with open(fn2, "r") as f_r:
    code2 = f_r.read()
m1 = ptn_version1.search(code1)
if not m1:
    raise KeyError(f"Could not find version number in {fn1}")
old_version = m1.group(1)
m2 = ptn_version2.search(code2)
if not m2:
    raise KeyError(f"Could not find version number in {fn2}")
print(f"OLD VERSION: {old_version}")

# work out next version number
new_version = old_version
if increment:
    new_version = incr_version(old_version)
print(f"NEW VERSION: {new_version}")

# modify code
code1 = replace_version(code1, m1, new_version)
code2 = replace_version(code2, m2, new_version)
if write:
    with open(fn1, "w") as f_w:
        f_w.write(code1)
    with open(fn2, "w") as f_w:
        f_w.write(code2)

# remove old distribution files
if write:
    for fn in glob.glob("dist/*"):
        os.remove(fn)

# build
if write:
    subprocess.run([sys.executable, "setup.py", "sdist", "bdist_wheel"], check=True)

# upload
#  -- prompts for username and password
if write:
    print("UPLOADING... (PyPi username & password will be needed)")
    subprocess.run(["twine", "upload", "dist/*"])

exit(0)



"""
An adapter for MMLibrary that simply delegates the calls to the existing (old) mmlibrary.
"""
import io
import json
import typing
from ..front import MMLibrary, MMException
from ..metadata import Metadata
# import the old mmlibrary
from .old_mm import *


class MMLibraryToOld(MMLibrary):
    def get_metadata(self, model_uri: str=None) -> Metadata:
        # TODO fill in model uri, at least
        # TODO fill in parameter metadata
        return Metadata()

    def get_arguments(self):
        raise MMException("not implemented")

    def get_argument(self, argument_name: str, default=None):
        try:
            get_argument(argument_name)
        except ValueError:
            return default

    def _open_reader_for_data(self, data: (str, bytes, bytearray), mode: str):
        if 'b' in mode:
            return io.BytesIO(data)
        else:
            return io.StringIO(data.decode("utf-8"))

    def _open_writer_with_save(self, save_binary, mode: str):
        if 'b' in mode:
            class WB(io.BytesIO):
                def close(self) -> None:
                    save_binary(self.getvalue())
            return WB()
        else:
            class WS(io.StringIO):
                def close(self) -> None:
                    save_binary(self.getvalue().encode("utf-8"))
            return WS()

    def open_resource(self, resource_name: str, mode: str='rb') -> io.IOBase:
        if 'r' in mode:
            bin = get_binary_from_resource(resource_name)
            return self._open_reader_for_data(bin, mode)
        else:
            return self._open_writer_with_save(lambda bin: save_binary_to_resource(resource_name, bin), mode)

    def open_run_data(self, filename: str, mode: str='rb') -> io.IOBase:
        if 'r' in mode:
            data = get_temporary_data()
            if isinstance(data, dict):
                part = data.get(filename)
                if not isinstance(part, (str, bytes, bytearray)):
                    part = json.dumps(part)
                return self._open_reader_for_data(part)
            return self._open_reader_for_data(b'', mode)
        data = get_temporary_data()
        if not isinstance(data, dict):
            data = {}
        def writer(part):
            data[filename] = json.loads(part)
            save_temporary_data(data)
        return self._open_writer_with_save(writer, mode)

    def open_model_binary(self, part: str=None, mode: str='rb') -> io.IOBase:
        return open(get_model(), mode=mode)

    def get_metrics(self, metric_name: str = None, uri: str = None, time_range: (list, tuple) = None, limit: int = None) -> typing.Iterable:
        raise MMException("not implemented")

    def save_metric(self, metric_name: str, metric_value) -> None:
        save_kpi(kpi_name=metric_name, kpi_value=metric_value)

    def get_last_metric(self, metric_name: str):
        return get_last_kpi(metric_name)

    def set_evaluation_result(self, success: bool, message: str = None) -> None:
        if not success:
            evaluation_result(message)

    def scoring_response(self, score) -> None:
        return_score(score)

    def is_scoring_mode(self):
        raise MMException("not implemented")

    def get_db_connection(self, connection_name: str, database: str = None):
        return get_db_connection(connection_name)

    def iter_scoring_requests(self, callback=None):
        raise MMException("not implemented")

    def run(self, code_uri: str, arguments: dict = None, callback=None):
        raise MMException("not implemented")



"""
Adapter to "mmlibrary".

The "mmlibrary" package can be installed locally to provide a default custom target.
"""
import io
import json
import inspect

from dslibrary.engine_intf import FileSystem
from dslibrary.front import DSLibraryException
from dslibrary.metadata import Metadata
from dslibrary.utils.file_utils import write_stream_with_read_on_close
from dslibrary import DSLibrary

try:
    import mmlibrary
except ImportError:
    mmlibrary = None


class DSLibraryViaMMLibrary(DSLibrary):
    def __init__(self, *args, _mm=None, **kwargs):
        """
        :param _mm:    You can reference a custom set of functions instead of the mmlibrary package.
        """
        self._mm = _mm or mmlibrary
        super(DSLibraryViaMMLibrary, self).__init__(*args, **kwargs)

    def get_metadata(self):
        return Metadata()

    def _find_method(self, *args, mandatory: bool=True):
        """
        Find a method that matches one of the names given in 'args'.
        """
        for method in args:
            if hasattr(self._mm, method):
                return getattr(self._mm, method)
        if mandatory:
            raise DSLibraryException(f"method not found: {args[0]}()")

    def _setup_code_paths(self):
        """
        If mmlibrary defines a method to setup code paths we call it first.
        """
        setup_code_paths = self._find_method("setup_code_paths", mandatory=False)
        if setup_code_paths:
            setup_code_paths()
        super(DSLibraryViaMMLibrary, self)._setup_code_paths()

    def get_parameters(self):
        if not self._mm:
            return {}
        get_arguments = self._find_method("get_arguments", "get_parameters", mandatory=False)
        if get_arguments:
            return self._mm.get_arguments()
        if hasattr(self._mm, "param_dictionary"):
            return self._mm.param_dictionary
        raise DSLibraryException(f"method not found: get_arguments()")

    def get_parameter(self, parameter_name: str, default=None):
        if not self._mm:
            return default
        method = self._find_method("get_parameter", "get_argument", "getArgument")
        try:
            return method(parameter_name)
        except (ValueError, KeyError):
            pass
        return default

    def _opener(self, path: str, mode: str, **kwargs) -> io.IOBase:
        if 'w' in mode:
            writer = self._find_method("save_binary_to_resource", "saveBinaryToResource")
            def finalize(fh):
                writer(path, fh.read())
            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=finalize)
        if "r" in mode:
            try:
                reader = self._find_method("get_binary_from_resource", "getBinaryFromResource")
                if mode == 'rb':
                    return io.BytesIO(reader(path))
                return io.StringIO(reader(path).decode('utf-8'))
            except ValueError:
                raise FileNotFoundError(f"not found: {path}")
        raise DSLibraryException(f"Unsupported mode: {mode}")

    def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
        """
        The model-binary is just one file, but we can support 'part' by assuming it is a ZIP file.
        """
        if part:
            import zipfile
            if mode in ("w", "wb", "a", "ab"):
                # we can't write parts because each call to mmlibrary.new_version() creates a new trained model
                raise ValueError(f"Unsupported mode for model binary with parts: {mode}")
            with self.open_model_binary() as f_r:
                with zipfile.ZipFile(f_r) as zf:
                    # TODO this won't work with large data -- write to a temporary file and read from there, delete file on close
                    if mode == "r":
                        return io.StringIO(zf.read(part).decode("utf-8"))
                    if mode == "rb":
                        return io.BytesIO(zf.read(part))
        if "r" in mode:
            get_model = self._find_method("get_model", "getModel")
            return open(get_model(), mode)
        if "w" in mode:
            wr_model = self._find_method("new_version", "newVersion")
            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=lambda f_r: wr_model(f_r.read()))

    def log_metric(self, metric_name: str, metric_value: float, step: int=0) -> None:
        """
        MMlibrary has its own KPI logging method.
        """
        if self._mlflow_metrics:
            return super(DSLibraryViaMMLibrary, self).log_metric(metric_name, metric_value, step)
        self._mm.save_kpi(metric_name, metric_value)

    def open_run_data(self, filename: str, mode: str='rb', **kwargs) -> io.IOBase:
        """
        The method provided by mmlibrary to save 'run data' only stores a single chunk of binary data.  We
        store JSON in that data, and translate calls to read and write files such that they read and write
        elements of that JSON object.
        """
        reader = self._find_method("get_temporary_data")
        if "r" in mode:
            data = json.loads(reader() or b'{}')
            if filename not in data:
                raise FileNotFoundError(f"not found: {filename}")
            if mode == 'r':
                return io.StringIO(data[filename])
            return io.BytesIO((data[filename]).encode("utf-8"))
        def finalize(fh):
            data = json.loads(reader() or b'{}')
            if "a" in mode:
                if filename not in data:
                    data[filename] = ""
                data[filename] += fh.read()
            else:
                data[filename] = fh.read()
            writer = self._find_method("save_temporary_data")
            writer(json.dumps(data).encode('utf-8'))
        return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=finalize)

    def get_next_scoring_request(self, timeout: float=None) -> (dict, None):
        raise DSLibraryException("not implemented in mmlibrary, use get_parameter() for each fields of a single scoring request")

    def scoring_response(self, score):
        method = self._find_method("return_score", "returnScore")
        method(score)

    def _get_system_connection(self, system_type, resource_name: str, for_write: bool=False, **kwargs):
        """
        All types of connection can be accessed through mmlibrary.get_db_connection().
        """
        # anything with a protocol can hopefully be handled by the built-in logic
        uri, _ = self._xlt_resource(resource_name, "outputs" if for_write else "inputs", **kwargs)
        if ":" in uri:
            return super(DSLibraryViaMMLibrary, self)._get_system_connection(system_type, resource_name, for_write, **kwargs)
        # ordinary 'named' connections go through this method to build a proxy to a remote system
        method = self._find_method("get_db_connection", mandatory=True)
        sig = inspect.signature(method)
        more = {}
        if "_expected_system_type" in sig.parameters:
            more["_expected_system_type"] = system_type
        if "_read_only" in sig.parameters:
            more["_read_only"] = not for_write
        db_or_bucket = kwargs.get("database") or kwargs.get("bucket_name")
        if db_or_bucket:
            if "_db_or_bucket" in sig.parameters:
                more["_db_or_bucket"] = db_or_bucket
            else:
                raise DSLibraryException("Incompatible use of 'database' or 'bucket_name' argument -- override is not supported by this version of mmlibrary")
        return method(resource_name, **more)

    def set_evaluation_result(self, success: bool, **kwargs) -> None:
        method = self._find_method("evaluation_result", mandatory=True)
        method("pass" if success else "fail")

    def install_packages(self, packages: list, verbose: bool=False):
        method = self._find_method("install_packages")
        sig = inspect.signature(method)
        if method:
            return method(packages)
        else:
            return super(DSLibraryViaMMLibrary, self).install_packages(packages, verbose)



"""
Complete 'file' implementations which read and write data in chunks.  Main intention is to make remote files accessed
via REST APIs look like local files.  Contains support for restartable uploads (i.e. on s3).
"""
import io
import inspect


class ChunkedFileReader(io.RawIOBase):
    """
    Makes a file-like read-only object from a function that can read chunks.
    """
    def __init__(self, filename: str, reader: callable, mode: str='r', chunk_size=16384, on_close: callable=None, metadata: dict=None, size=None):
        """
        :param filename:    Name to show in the 'name' property.
        :param reader:      Method which accepts a byte range.
        :param mode:        'r' or 'rb'.
        :param metadata:    Additional information to associate with the file (i.e. custom extensions).
        :param size:        Size of file, if known, or None, or a callable.
        """
        self.name = filename
        self.metadata = metadata
        self._size = size
        self._mode = mode
        self._reader = reader
        self._on_close = on_close
        self._pos = 0
        self._offset = 0
        self._chunk = None
        self._eof = False
        self._eol = b'\n' if 'b' in mode else '\n'
        self._min_chunk = chunk_size
        self._max_read = 1048576*1024*1024
        self._output_type = bytearray if 'b' in self._mode else str
        self._is_binary = 'b' in self._mode

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self):
        return True

    def __getitem__(self, item):
        if not isinstance(item, slice):
            item = slice(item, item+1)
        # limit to a particular line range
        return slicer(self, item)

    def _fix_chunk(self, chunk):
        if 'b' in self._mode:
            if isinstance(chunk, str):
                # str -> bytes
                chunk = chunk.encode("utf-8")
        else:
            if isinstance(chunk, bytes):
                # bytes -> str
                chunk = chunk.decode("utf-8")
        return chunk

    def _get_size(self):
        if self._size is None:
            return
        if hasattr(self._size, "__call__"):
            self._size = self._size()
        return self._size

    def __iter__(self):
        """
        Read lines.
        """
        flush = self._chunk is None
        out = self._output_type()
        while True:
            if flush:
                self._chunk = self._fix_chunk(self._reader(self._pos, self._pos + self._min_chunk))
                self._offset = 0
                self._eof = self._eof or not self._chunk or len(self._chunk) < self._min_chunk
                flush = False
            try:
                eol2 = self._chunk.index(self._eol, self._offset)
                out += self._chunk[self._offset:eol2]
                eol2 += len(self._eol)
                self._pos += eol2 - self._offset
                self._offset = eol2
                yield out
                out = self._output_type()
            except ValueError:
                # EOL not found in chunk
                out += self._chunk[self._offset:]
                if self._eof:
                    if out:
                        yield out
                    break
                self._pos += len(self._chunk) - self._offset
                self._offset = len(self._chunk)
                flush = True

    def _read_direct(self, n):
        self._offset = 0
        self._chunk = None
        out = self._fix_chunk(self._reader(self._pos, self._pos + n))
        self._pos += len(out)
        if len(out) < n:
            self._eof = True
        return out

    def read(self, n=None):
        if not n or n > self._max_read:
            return self.read(self._max_read)
        # read large chunks directly, no buffer
        if n >= self._min_chunk*2 and not self._eof and not self._offset:
            return self._read_direct(n)
        out = self._output_type()
        while True:
            # make sure we have a chunk
            if (self._chunk is None or self._offset >= len(self._chunk)) and not self._eof:
                self._chunk = self._fix_chunk(self._reader(self._pos, self._pos + self._min_chunk))
                self._eof = not self._chunk or len(self._chunk) < self._min_chunk
                self._offset = 0
            if not self._chunk:
                break
            # take what we can from this chunk
            part = self._chunk[self._offset: self._offset + n]
            out += part
            n -= len(part)
            self._pos += len(part)
            self._offset += len(part)
            if n <= 0 or self._eof:
                break
        if self._is_binary:
            # NOTE: some code (i.e. pandas.read_csv()) assumes 'bytes' and fails on 'bytearray'
            return bytes(out)
        return out

    def tell(self):
        return self._pos

    def seek(self, n: int, whence: int=None):
        if whence == 2:
            if self._eof:
                if n >= 0:
                    return
                n += self._pos
            else:
                size = self._get_size()
                if size is None:
                    raise Exception("not supported, seek(*, 2) except at EOF")
                n += size
            whence = 0
        if whence == 1:
            n += self._pos
        if self._chunk is not None:
            buf0 = self._pos - self._offset
            buf_n = buf0 + len(self._chunk)
            if buf0 <= n < buf_n:
                # seek is within current chunk
                o2 = n - buf0
                self._pos += o2 - self._offset
                self._offset = o2
                self._eof = False
                return
        self._pos = n
        self._chunk = None
        self._eof = False

    def close(self):
        if self._on_close:
            self._on_close()


class ChunkedFileWriter(object):
    """
    Makes a file-like writable object from a function that can write/append chunks.

    The writer() method takes 2-3 arguments: the data to write, an 'append' flag, and then optionally a 'hint'
    flag.

    The normal/simple implementation only defines the first two flags.  Assuming only one process is
    updating a given file at one time, and assuming the requests arrive in order, appending works for
    transmission of large files.  In this implementation there is no return value from writer().

    The 'hint' flag is required to bypass those limitations.  Here is how a multi-part upload works using such
    an implementation:
      upload_id = writer(FIRST_CHUNK, False, hint="start:TOTAL_FILE_SIZE")
      writer(SECOND_CHUNK, True, hint="continue:UPLOAD_ID:1")
      writer(...)
      writer(LAST_CHUNK, True, hint="end:UPLOAD_ID:N")
    """
    def __init__(self, filename: str, writer: callable, mode: str='w', chunk_size=10000000, metadata: dict=None, total_size: int=None):
        """
        :param filename:    Name to show in the 'name' property.
        :param writer:      Method which writes data of the type indicated by mode.
        :param mode:        'w' or 'wb'.
        :param chunk_size:  How much data to write at once.  Note that for S3 the minimum is 5Mi.
        :param metadata:    Additional information to associate with the file (i.e. custom extensions).
        :param total_size:  Total size of data being transferred, if known.
        """
        self.name = filename
        self.metadata = metadata
        self._mode = mode
        self._writer = writer
        self._can_hint = "hint" in inspect.signature(writer).parameters.keys()
        self._append = 'a' in mode
        self._buffer = bytearray() if 'b' in mode else ""
        self._chunk_size = chunk_size
        self._chunk_no = 0
        self._mpu_id = None
        self._total_size = total_size

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def readable(self) -> bool:
        return False

    def writable(self) -> bool:
        return True

    def seekable(self):
        return False

    def write(self, content):
        if isinstance(content, str) == ('b' in self._mode):
            raise TypeError("invalid data type for mode=%s" % self._mode)
        while len(self._buffer) + len(content) >= self._chunk_size:
            part = self._chunk_size - len(self._buffer)
            self._buffer += content[:part]
            self.flush()
            content = content[part:]
        self._buffer += content

    def flush(self, last: bool=False):
        if self._buffer or last:
            self._chunk_no += 1
            single_chunk = self._chunk_no == 1 and last
            if self._can_hint and not single_chunk:
                hint, phase = self._hint_phase(last)
                resp = self._writer(self._buffer, self._append, hint=hint)
                if phase == "start":
                    self._mpu_id = resp
            else:
                self._writer(self._buffer, self._append)
            self._append = True
            self._buffer = bytearray() if 'b' in self._mode else ""

    def _hint_phase(self, last: bool):
        if self._chunk_no == 1:
            phase = "start"
        else:
            phase = "end" if last else "continue"
        if phase == "start":
            hint = "start:%d" % (self._total_size or -1)
        elif self._mpu_id:
            hint = "%s:%s:%d" % (phase, self._mpu_id, self._chunk_no - 1)
        else:
            hint = None
        return hint, phase

    def close(self):
        self.flush(last=True)


def slicer(iterable, which):
    """
    Supports sliced iteration for another iterator.
    """
    if not isinstance(which, slice):
        which = slice(which, which+1)
    # reject negative start/stop/step values
    if \
            (which.start is not None and which.start < 0) or \
            (which.stop is not None and which.stop < 0) or \
            (which.step is not None and which.step < 0):
        raise Exception("not allowed: negative slice values")
    # generate an iterable
    return SlicerFetcher(iterable, which)


class SlicerFetcher(object):
    def __init__(self, iterable, which):
        self.iterable = iterable
        self.which = which
        self.pos = -1
        self.item = None

    def __iter__(self):
        i = iter(self.iterable)
        def skip():
            if not self.which.step or self.which.step < 2:
                return
            for _ in range(self.which.step - 1):
                next(i)
        try:
            if self.which.start:
                for _ in range(self.which.start):
                    next(i)
            if self.which.stop is None:
                while True:
                    yield next(i)
                    skip()
            n = self.which.stop - (self.which.start or 0)
            if self.which.step:
                n /= self.which.step
            for _ in range(n):
                yield next(i)
                skip()
        except StopIteration:
            pass
